﻿using System;

namespace KnightTour
{
    public class WarnsdorffAlgorithm
    {
        public WarnsdorffAlgorithm()
        {
            // Create board.
            board = new int[8][];
            for (int i = 0; i < 8; i++)
            {
                board[i] = new int[8];
            }
        }

        public int[][] Solve(int positionX, int positionY)
        {
            // Check argument(s).
            if ((positionX < 0) || (positionX >= 8))
            {
                throw new ArgumentOutOfRangeException("positionX");
            }
            if ((positionY < 0) || (positionY >= 8))
            {
                throw new ArgumentOutOfRangeException("positionY");
            }

            ClearBoard();

            int moveNumber = 1;
            board[positionX][positionY] = moveNumber;
            moveNumber++;

            for (int i = 1; i < 64; i++)
            {
                bool moved = GetNextMoves(ref positionX, ref positionY);

                if (!moved)
                {
                    return null;
                }

                board[positionX][positionY] = moveNumber;
                moveNumber++;
            }

            return (int[][])board.Clone();
        }

        private int[][] board;

        private void ClearBoard()
        {
            for (int positionX = 0; positionX < 8; positionX++)
            {
                for (int positionY = 0; positionY < 8; positionY++)
                {
                    board[positionX][positionY] = 0;
                }
            }
            return;
        }

        private bool GetNextMoves(ref int positionX, ref int positionY)
        {
            int moveX = -1;
            int moveY = -1;
            bool moved = false;
            int accessibility = 8;

            // (X + 2, Y + 1)
            if (((positionX + 2) < 8) && ((positionY + 1) < 8) && (board[positionX + 2][positionY + 1] == 0))
            {
                if ((GetAccessibility((positionX + 2), (positionY + 1)) < accessibility))
                {
                    accessibility = GetAccessibility((positionX + 2), (positionY + 1));
                    moveX = positionX + 2;
                    moveY = positionY + 1;
                    moved = true;
                }
            }

            // (X + 2, Y - 1)
            if (((positionX + 2) < 8) && ((positionY - 1) >= 0) && (board[positionX + 2][positionY - 1] == 0))
            {
                if ((GetAccessibility((positionX + 2), (positionY - 1)) < accessibility))
                {
                    accessibility = GetAccessibility((positionX + 2), (positionY - 1));
                    moveX = positionX + 2;
                    moveY = positionY - 1;
                    moved = true;
                }

            }

            // (X + 1, Y + 2)
            if (((positionX + 1) < 8) && ((positionY + 2) < 8) && (board[positionX + 1][positionY + 2] == 0))
            {
                if ((GetAccessibility((positionX + 1), (positionY + 2)) < accessibility))
                {
                    accessibility = GetAccessibility((positionX + 1), (positionY + 2));
                    moveX = positionX + 1;
                    moveY = positionY + 2;
                    moved = true;
                }

            }

            // (X + 1, Y - 2)
            if (((positionX + 1) < 8) && ((positionY - 2) >= 0) && (board[positionX + 1][positionY - 2] == 0))
            {
                if ((GetAccessibility((positionX + 1), (positionY - 2)) < accessibility))
                {
                    accessibility = GetAccessibility((positionX + 1), (positionY - 2));
                    moveX = positionX + 1;
                    moveY = positionY - 2;
                    moved = true;
                }

            }

            // (X - 1, Y + 2)
            if (((positionX - 1) >= 0) && ((positionY + 2) < 8) && (board[positionX - 1][positionY + 2] == 0))
            {
                if ((GetAccessibility((positionX - 1), (positionY + 2)) < accessibility))
                {
                    accessibility = GetAccessibility((positionX - 1), (positionY + 2));
                    moveX = positionX - 1;
                    moveY = positionY + 2;
                    moved = true;
                }

            }

            // (X - 1, Y - 2)
            if (((positionX - 1) >= 0) && ((positionY - 2) >= 0) && (board[positionX - 1][positionY - 2] == 0))
            {
                if ((GetAccessibility((positionX - 1), (positionY - 2)) < accessibility))
                {
                    accessibility = GetAccessibility((positionX - 1), (positionY - 2));
                    moveX = positionX - 1;
                    moveY = positionY - 2;
                    moved = true;
                }

            }

            // (X - 2, Y + 1)
            if (((positionX - 2) >= 0) && ((positionY + 1) < 8) && (board[positionX - 2][positionY + 1] == 0))
            {
                if ((GetAccessibility((positionX - 2), (positionY + 1)) < accessibility))
                {
                    accessibility = GetAccessibility((positionX - 2), (positionY + 1));
                    moveX = positionX - 2;
                    moveY = positionY + 1;
                    moved = true;
                }

            }

            // (X - 2, Y - 1)
            if (((positionX - 2) >= 0) && ((positionY - 1) >= 0) && (board[positionX - 2][positionY - 1] == 0))
            {
                if ((GetAccessibility((positionX - 2), (positionY - 1)) < accessibility))
                {
                    accessibility = GetAccessibility((positionX - 2), (positionY - 1));
                    moveX = positionX - 2;
                    moveY = positionY - 1;
                    moved = true;
                }

            }

            positionX = moveX;
            positionY = moveY;
            return moved;
        }

        private int GetAccessibility(int positionX, int positionY)
        {
            int accessibility = 0;

            // (X + 2, Y + 1)
            if (((positionX + 2) < 8) && ((positionY + 1) < 8) && (board[positionX + 2][positionY + 1] == 0))
            {
                accessibility++;
            }

            // (X + 2, Y - 1)
            if (((positionX + 2) < 8) && ((positionY - 1) >= 0) && (board[positionX + 2][positionY - 1] == 0))
            {
                accessibility++;
            }

            // (X + 1, Y + 2)
            if (((positionX + 1) < 8) && ((positionY + 2) < 8) && (board[positionX + 1][positionY + 2] == 0))
            {
                accessibility++;
            }

            // (X + 1, Y - 2)
            if (((positionX + 1) < 8) && ((positionY - 2) >= 0) && (board[positionX + 1][positionY - 2] == 0))
            {
                accessibility++;
            }

            // (X - 1, Y + 2)
            if (((positionX - 1) >= 0) && ((positionY + 2) < 8) && (board[positionX - 1][positionY + 2] == 0))
            {
                accessibility++;
            }

            // (X - 1, Y - 2)
            if (((positionX - 1) >= 0) && ((positionY - 2) >= 0) && (board[positionX - 1][positionY - 2] == 0))
            {
                accessibility++;
            }

            // (X - 2, Y + 1)
            if (((positionX - 2) >= 0) && ((positionY + 1) < 8) && (board[positionX - 2][positionY + 1] == 0))
            {
                accessibility++;
            }

            // (X - 2, Y - 1)
            if (((positionX - 2) >= 0) && ((positionY - 1) >= 0) && (board[positionX - 2][positionY - 1] == 0))
            {
                accessibility++;
            }
            return accessibility;
        }
    }
}
